#ifndef VELOCITY_FIELDS_H
#define VELOCITY_FIELDS_H

#include <Core/Log/DEBUG_UTILITIES.h>
#include <Core/Vectors/VECTOR.h>
#include <cmath>

namespace PhysBAM{

template<class TV> using FIELD=TV(*)(const TV&X);
template<class TV> using FIELD_AVG=TV(*)(const TV&X,typename TV::SCALAR dx);
template<class TV> using BOTH=std::pair<FIELD<TV>,FIELD_AVG<TV>>;

// fields 0-3 are for div free
// fields 4-7 are for curl free
template<class TV>
BOTH<TV> Velocity_Field_Both(int i)
{
    typedef typename TV::SCALAR T;
    typedef TV (*FUNC)(const TV&X,T dx);

    if constexpr(TV::m == 2)
    {
        static BOTH<TV> list[]=
        {
            // disc div free rng
            {[](const TV&X){T x=X.x,y=X.y;return TV(sin(370.*x + 2) * sin(370*y + 4), cos(370*x + 2)*cos(370*y + 4));},
             [](const TV&X, T dx){T x=X.x,y=X.y; return TV(-1./370*(cos(370.*dx + 4) - cos(370.*dx - 4))*sin(370.*x + 2), 1./370*(sin(370*dx + 2) + sin(370*dx - 2))*cos(370*y + 4));}},
            // disc div free
            {[](const TV&X){T x=X.x,y=X.y;return TV(sin(x + 2)*sin(y + 4), cos(x + 2)*cos(y + 4));},
             [](const TV&X,T dx){T x=X.x,y=X.y;return TV(sin(x+2)*(cos(-y+1./2*dx-4)-cos(y+1./2*dx+4)),cos(y+4)*(sin(-x+1./2*dx-2)+sin(x+1./2*dx+2)))/dx;}},
            // analytically div free, not disc
            // disc div is O(dx^2)
            {[](const TV&X){T x=X.x,y=X.y;return TV(x*x*x - 6*x*y*y + y*y*y, -3*x*x*y + 2*y*y*y);},{}},
            // neither
            {[](const TV&X){T x=X.x,y=X.y;return TV(sin(x + 2) + sin(y + 4), cos(x + 2) + cos(y + 4));},{}},
            // disc curl free rng
            {[](const TV&X){T x=X.x,y=X.y;return TV(sin(370.*x + 2) * cos(370*y + 4), cos(370*x + 2)*sin(370*y + 4));},{}},
            // disc curl free
            {[](const TV&X){T x=X.x,y=X.y;return TV(sin(x + 2) * cos(y + 4), cos(x + 2)*sin(y + 4));},{}},
            // analytically curl free, not disc
            // grad -x^3*y+x*y^3
            // note that the discrete curl is still O(dx^2)
            {[](const TV&X){T x=X.x,y=X.y;return TV(-3*x*x*y+y*y*y, -x*x*x+3*x*y*y);},{}},
        };
        PHYSBAM_ASSERT(i<sizeof(list)/sizeof(*list));
        return list[i];
    }
    if constexpr(TV::m == 3)
    {
        static BOTH<TV> list[]=
        {
            {[](const TV&X){T x=X.x,y=X.y,z=X.z;return TV(
                    sin(370*x + 2)*sin(370*y + 4) *sin(370*z + 6),
                    cos(370*x + 2)*cos(370*y + 4)* cos(370*z + 6),
                    cos(370*x + 2)*sin(370*y + 4)*(cos(370*z + 6) + sin(370*z + 6)));},
             [](const TV& X, T dx){T x=X.x,y=X.y,z=X.z;return TV(
                     1./136900*(cos(370*dx + 6) - cos(370*dx - 6))*(cos(370*dx + 4) - cos(370*dx - 4))*sin(370*x + 2),
                     1./136900*(sin(370*dx + 6) + sin(370*dx - 6))*(sin(370*dx + 2) + sin(370*dx - 2))*cos(370*y + 4),
                     -1./136900*(cos(370*dx + 4) - cos(370*dx - 4))*(cos(370*z + 6) + sin(370*z + 6))*(sin(370*dx + 2) + sin(370*dx - 2)));}
            },
            {[](const TV&X){T x=X.x,y=X.y,z=X.z;return TV(
                        sin(x + 2)*sin(y + 4)*sin(z + 6),
                        cos(x + 2)*cos(y + 4)*cos(z + 6),
                        cos(x + 2)*sin(y + 4)*(cos(z + 6)+sin(z + 6)));},
             [](const TV&X,T dx){T x=X.x,y=X.y,z=X.z;return TV(
                     sin(x+2)*(cos(-y+1./2*dx-4)-cos(y+1./2*dx+4))*(-cos(6+z+1./2*dx)+cos(-6-z+1./2*dx)),
                     cos(y+4)*(sin(-x+1./2*dx-2)+sin(x+1./2*dx+2))*(sin(6+z+1./2*dx)+sin(-6-z+1./2*dx)),
                     (cos(-y+1./2*dx-4)-cos(y+1./2*dx+4))*(cos(z+6)+sin(z+6))*(sin(-x+1./2*dx-2)+sin(x+1./2*dx+2)))/(dx*dx);}},
            {[](const TV&X){T x=X.x,y=X.y,z=X.z;return TV(
                        y*y*y*z*z - 6*x*y*y*z + x*x*x,
                        3*x*x*y + 2*x*y*z + z*z,
                        3*y*y*z*z - 6*x*x*z - x*z*z);},{}},
            {[](const TV&X){T x=X.x,y=X.y,z=X.z;return TV(
                        sin(x + 2) + sin(y + 4),
                        cos(y + 2) + cos(z + 4),
                        cos(z + 2) + cos(x + 4));},{}},
            // disc curl free rng
            {[](const TV&X){T x=X.x,y=X.y,z=X.z;return TV(
                        sin(370*x + 2)*cos(370*y + 4)*cos(370*z + 6),
                        cos(370*x + 2)*sin(370*y + 4)*cos(370*z + 6),
                        cos(370*x + 2)*cos(370*y + 4)*sin(370*z + 6));},{}},
            // disc curl free
            {[](const TV&X){T x=X.x,y=X.y,z=X.z;return TV(
                        sin(x + 2)*cos(y + 4)*cos(z + 6),
                        cos(x + 2)*sin(y + 4)*cos(z + 6),
                        cos(x + 2)*cos(y + 4)*sin(z + 6));},{}},
            // analytically curl free, not disc
            // grad -x^3*y+x*y^3
            // note that the discrete curl is still O(dx^2)
            {[](const TV&X){T x=X.x,y=X.y;return TV(
                        -3*x*x*y+y*y*y,
                        -x*x*x+3*x*y*y,
                        0);},{}},
        };
        PHYSBAM_ASSERT(i<sizeof(list)/sizeof(*list));
        return list[i];
    }
}

template<class TV>
FIELD<TV> Velocity_Field(int i)
{return Velocity_Field_Both<TV>(i).first;}
template<class TV>
FIELD_AVG<TV> Velocity_Field_Average(int i)
{return Velocity_Field_Both<TV>(i).second;}
}

#endif

