#include <Core/Arrays/ARRAY.h>
#include <Core/Grids/FACE_INDEX.h>
#include <Core/Grids/FACE_ITERATOR.h>
#include <Core/Grids/GRID.h>
#include <Core/Grids/NODE_ITERATOR.h>
#include <Core/Interpolation/CUBIC_INTERPOLATION_FACE.h>
#include <Core/Interpolation/CUBIC_SPLINE_INTERPOLATION_FACE.h>
#include <Core/Interpolation/CURL_FREE_INTERPOLATION.h>
#include <Core/Interpolation/CURL_FREE_INTERPOLATION_4TH_ORDER.h>
#include <Core/Interpolation/CURL_FREE_INTERPOLATION_C1.h>
#include <Core/Interpolation/DIV_FREE_BALSARA.h>
#include <Core/Interpolation/DIV_FREE_INTERPOLATION.h>
#include <Core/Interpolation/DIV_FREE_INTERPOLATION_4TH_ORDER.h>
#include <Core/Interpolation/DIV_FREE_INTERPOLATION_C1.h>
#include <Core/Interpolation/DIV_FREE_INTERPOLATION_C1_INTERP.h>
#include <Core/Interpolation/DIV_FREE_INTERPOLATION_FLUXES.h>
#include <Core/Interpolation/DIV_FREE_INTERPOLATION_INTERP.h>
#include <Core/Interpolation/GENERAL_INTERPOLATION_FACE.h>
#include <Core/Interpolation/LINEAR_INTERPOLATION_FACE.h>
#include <Core/Log/LOG.h>
#include <Core/Matrices/MATRIX.h>
#include <Core/Parsing/PARSE_ARGS.h>
#include <Core/Random_Numbers/RANDOM_NUMBERS.h>
#include <Core/Vectors/VECTOR.h>
#include "PROJECT_FFT.h"
#include "VELOCITY_FIELDS.h"
#include <array>
#include <fstream>

using namespace PhysBAM;

typedef double RW;
typedef double T;

T eps=1e-6;

//constexpr static int num_positions = 1'000'000;
constexpr static int num_positions = 1'000'000;
constexpr static int num_schemes_div = 10;
constexpr static int num_schemes_curl = 6;
constexpr static int ghost = 3;

ARRAY<T> ST2={-1,1};
ARRAY<T> ST4={1/24.,-27/24.,27/24.,-1/24.};


template<typename TV>
VECTOR<T, num_schemes_curl> Run_Test_Curl(const GRID<TV>& grid, const ARRAY<T, FACE_INDEX<TV::m>>& velocity)
{
    typedef VECTOR<int, TV::m> IV;

    RANDOM_NUMBERS<T> rand(123);

    ARRAY<TV> positions(num_positions);
    rand.Fill_Uniform(positions, grid.Domain());

    for(auto& X:positions)
    {
        while((grid.Cell_X(grid.Cell_I(X))-X).Max_Abs()>(T).5*grid.dX(0)-2*eps ||
            (grid.Node_X(grid.Node_I(X))-X).Max_Abs()>(T).5*grid.dX(0)-2*eps)
            rand.Fill_Uniform(X,TV(),TV()+1);
    }

    LINEAR_INTERPOLATION_FACE<TV,T> i0;
    CUBIC_INTERPOLATION_FACE<TV,T> i1;
    CUBIC_SPLINE_INTERPOLATION_FACE<TV,T> i2;
    CURL_FREE_INTERPOLATION<TV, T> i3;
    CURL_FREE_INTERPOLATION_C1<TV, T> i4;
    CURL_FREE_INTERPOLATION_4TH_ORDER<TV, T> i5;

    std::array<INTERPOLATION_FACE<TV, T>*, num_schemes_curl> interps{&i0 ,&i1, &i2, &i3, &i4, &i5};
    std::array<std::string, num_schemes_curl> names{"linear", "cubic", "cubic spline", "curl-free-c0", "curl-free-c1", "curl-free-4th"};

    VECTOR<T, num_schemes_curl> curl;

#pragma omp parallel for
    for(int i = 0; i < interps.size(); i++) 
    {
        T max_curl=0;
        auto interp=interps[i];
        for(int p=0;p<positions.m;p++){
            // partial_(d) u_a
            auto diff=[&](int d,int a){
                TV X=positions(p);
                TV X1=X.Add_Axis(d,eps);
                TV X2=X.Add_Axis(d,-eps);
                return (interp->Interpolate_Face(grid, velocity, X1)(a) 
                    - interp->Interpolate_Face(grid, velocity, X2)(a))/(2*eps);
            };

            typename TV::SPIN c;
            if constexpr(TV::m==2){
                c={diff(0,1)-diff(1, 0)};
            } else if constexpr(TV::m==3){
                c=TV(diff(1,2)-diff(2,1),
                        diff(2,0)-diff(0,2),
                        diff(0,1)-diff(1,0));
            }
            T mag=c.Magnitude();
            //LOG::printf("%p at %p\n", mag, positions(p));
            if (mag>max_curl) max_curl=mag;
        }
        LOG::printf("%p %p\n", names[i], max_curl);
        curl[i] = max_curl;
    }
    return curl;
}


template<typename TV>
VECTOR<T, num_schemes_div> Run_Test_Div(const GRID<TV>& grid,const ARRAY<T, FACE_INDEX<TV::m>>& velocity)
{
    typedef VECTOR<int, TV::m> IV;

    RANDOM_NUMBERS<T> rand(123);

    ARRAY<TV> positions(num_positions);
    rand.Fill_Uniform(positions, grid.Domain());

    for(auto& X:positions)
    {
        while((grid.Cell_X(grid.Cell_I(X))-X).Max_Abs()>(T).5*grid.dX(0)-2*eps ||
            (grid.Node_X(grid.Node_I(X))-X).Max_Abs()>(T).5*grid.dX(0)-2*eps)
            rand.Fill_Uniform(X,TV(),TV()+1);
    }

    LINEAR_INTERPOLATION_FACE<TV,T> i1;
    CUBIC_INTERPOLATION_FACE<TV,T> i2;
    CUBIC_SPLINE_INTERPOLATION_FACE<TV,T> i3;
    DIV_FREE_BALSARA<TV,T> i4;
    DIV_FREE_INTERPOLATION<TV,T> i5;
    DIV_FREE_INTERPOLATION_INTERP<TV,T> i6;
    DIV_FREE_INTERPOLATION_C1<TV,T> i7;
    DIV_FREE_INTERPOLATION_C1_INTERP<TV,T> i8;
    DIV_FREE_INTERPOLATION_4TH_ORDER<TV,T> i9;
    DIV_FREE_INTERPOLATION_FLUXES<TV,T> i10;

    std::array<INTERPOLATION_FACE<TV, T>*, num_schemes_div> interps{&i1, &i2, &i3, &i4, &i5, &i6, &i7, &i8, &i9, &i10};
    std::array<std::string, num_schemes_div> names{"linear", "cubic", "cubic-spline", "balsara", "div-free-c0", "div-free-c0-interp", "div-free-c1", "div-free-c1-interp", "div-free-4th", "div-free-flux"};
    VECTOR<T, num_schemes_div> divergence;

#pragma omp parallel for
    for(int i = 0; i < interps.size(); i++) 
    {
        INTERPOLATION_FACE<TV, T>* interp = interps[i];
        auto interp_func = [interp,&grid,&velocity](TV x) 
        {
            return interp->Interpolate_Face(grid, velocity, x);
        };
        T max_div = 0;
        T max_diff = 0;
        T l2_div = 0;

        ARRAY<std::pair<TV,T> > vals;
        for(int p = 0; p < positions.m; p++) 
        {
            TV X = positions(p);
            T div = 0;
            for(int a = 0; a < TV::m; a++) 
            {
                TV A = X.Add_Axis(a, eps);
                TV B = X.Add_Axis(a, -eps);
                PHYSBAM_ASSERT(grid.Cell_I(A)==grid.Cell_I(B));
                T diff=(interp_func(A)(a) - interp_func(B)(a)) / (2 * eps);
                div += diff;
                max_diff = max(max_diff,abs(diff));
            }

            div=abs(div);
            vals.Append({X,div});
            max_div = max(max_div, div);
            l2_div += sqr(div);
        }

        l2_div=sqrt(l2_div/positions.m);
        divergence[i] = max_div;
        LOG::printf("%P %P %P %P\n",names[i],max_div,l2_div,max_diff);
    }

    return divergence;
}

template <class TV>
void Generate_Projected_Field_Div(const GRID<TV>& grid, ARRAY<T,FACE_INDEX<TV::m>>& vel, const ARRAY<T>& ST)
{

    RANDOM_NUMBERS<T> rand(123);

    for(FACE_ITERATOR<TV> it(grid, ghost);it.Valid();it.Next()){
        vel(it.Full_Index()) = rand.Get_Number();
    }

    Project_Div<T,TV::m,TV>(grid,vel, ST);
    Test_Div(vel,grid.numbers_of_cells,grid,ST);
}

template <class TV>
void Generate_Projected_Field_Curl(const GRID<TV>& grid, ARRAY<T,FACE_INDEX<TV::m>>& vel, const ARRAY<T>& ST)
{

    RANDOM_NUMBERS<T> rand(123);

    for(FACE_ITERATOR<TV> it(grid, ghost);it.Valid();it.Next()){
        vel(it.Full_Index()) = rand.Get_Number();
    }

    Project_Curl<T,TV::m,TV>(grid,vel, ST);
    Test_Curl(vel,grid.numbers_of_cells,grid,ST);
}


template<class TV>
void Do_Dim()
{
    typedef VECTOR<int, TV::m> IV;
    
    int res = 16;
    GRID<TV> grid(IV()+res, RANGE<TV>::Unit_Box());

    MATRIX<T, num_schemes_div, 6> divergence;
    ARRAY<T, FACE_INDEX<TV::m>> velocity(grid.Cell_Indices(ghost));

    for (int i=0; i < 4; i++) 
    {
        LOG::cout << " -------- VELOCITY FIELD: " << i  << " ----------"<< std::endl;

        auto field=Velocity_Field<TV>(i);
        for(FACE_ITERATOR it(grid,ghost); it.Valid(); it.Next()) 
        {
            auto fi = it.Full_Index();
            TV X = it.Location();

            velocity(fi) = field(X)(fi.axis);
        }

        divergence.Set_Column(i, Run_Test_Div<TV>(grid, velocity));
    }

    LOG::cout << " -------- VELOCITY FIELD: 2nd order projected ----------"<< std::endl;
    Generate_Projected_Field_Div(grid, velocity, ST2);
    divergence.Set_Column(4, Run_Test_Div<TV>(grid, velocity));

    LOG::cout << " -------- VELOCITY FIELD: 4th order projected ----------"<< std::endl;
    Generate_Projected_Field_Div(grid, velocity, ST4);
    divergence.Set_Column(5, Run_Test_Div<TV>(grid, velocity));


    static const char* latex_names_div[] =
    {
        "$B^0$","Cubic","$B^3$","Balsara","$C^0$","$C^{0i}$","$C^1$","$C^{1i}$", "4-th order", "Flux Preserving"
    };

    LOG::printf(" ---- DIV %PD ----\n",TV::m);
    for(int i = 0; i < divergence.Rows(); i++) 
    {
        LOG::printf("\\hline %8P",latex_names_div[i]);
        for(int j = 0; j < divergence.Columns(); j++) 
            LOG::printf("& $\\num{%.2e}$ ",divergence(i, j));
        LOG::printf("\\\\\n");
    }

    static const char* latex_names_curl[] =
    {
        "$B^0$","Cubic","$B^3$", "$C^0$ Curl", "$C^1$ Curl", "4th order"
    };
    MATRIX<T, num_schemes_curl, 6> curl;
    LOG::printf(" -------------------- CURL FREE TESTS ------------- \n");
    for (int i=0; i < 4; i++){
        int vel=i<3?i+4:i;
        LOG::cout << " -------- VELOCITY FIELD: " << vel  << " ----------"<< std::endl;
        auto field=Velocity_Field<TV>(vel);
        for(FACE_ITERATOR it(grid,ghost); it.Valid(); it.Next()) 
        {
            auto fi = it.Full_Index();
            TV X = it.Location();
            velocity(fi) = field(X)(fi.axis);
        }


        Test_Curl(velocity,grid.numbers_of_cells,grid,ST2);
        curl.Set_Column(i, Run_Test_Curl(grid,velocity));
    }

    LOG::cout << " -------- VELOCITY FIELD: 2nd order projected (curl) ----------"<< std::endl;
    Generate_Projected_Field_Curl(grid, velocity, ST2);
    curl.Set_Column(4, Run_Test_Curl<TV>(grid, velocity));

    LOG::cout << " -------- VELOCITY FIELD: 4th order projected (curl) ----------"<< std::endl;
    Generate_Projected_Field_Curl(grid, velocity, ST4);
    curl.Set_Column(5, Run_Test_Curl<TV>(grid, velocity));

    LOG::printf(" ---- CURL %PD ----\n",TV::m);
    for(int i = 0; i < curl.Rows(); i++) 
    {
        LOG::printf("\\hline %8P",latex_names_curl[i]);
        for(int j = 0; j < curl.Columns(); j++) 
            LOG::printf("& $\\num{%.2e}$ ",curl(i, j));
        LOG::printf("\\\\\n");
    }

}

int main(int argc, char* argv[])
{
    PARSE_ARGS parse_args(argc,argv);
    int dim=2;
    parse_args.Add("-dim",&dim,"dim","dimension to test");
    parse_args.Parse();
    Do_Dim<VECTOR<double,2>>();
    Do_Dim<VECTOR<double,3>>();
    return 0;
}
