#include <Core/Arrays/ARRAY.h>
#include <Core/Grids/FACE_ITERATOR.h>
#include <Core/Grids/GRID.h>
#include <Core/Interpolation/CUBIC_INTERPOLATION.h>
#include <Core/Interpolation/CUBIC_SPLINE_INTERPOLATION.h>
#include <Core/Interpolation/DIV_FREE_BALSARA.h>
#include <Core/Interpolation/DIV_FREE_INTERPOLATION.h>
#include <Core/Interpolation/DIV_FREE_INTERPOLATION_FLUXES.h>
#include <Core/Interpolation/LINEAR_INTERPOLATION_FACE.h>
#include <Core/Math_Tools/GAUSS_QUADRATURE.h>
#include <Core/Math_Tools/Quadrature_Integration.h>
#include <Core/Parsing/PARSE_ARGS.h>
#include "VELOCITY_FIELDS.h"
#include <array>

using namespace PhysBAM;

typedef double RW;
typedef double T;

template<class TV>
T Compute_Max_Flux_Error(GRID<TV> grid, ARRAY<T, FACE_INDEX<TV::m>> velocities, INTERPOLATION_FACE<TV,T>& interp) {
    T max_err=0;

    // -1 avoids boundary cells
    for(FACE_ITERATOR<TV> it(grid,-1);it.Valid();it.Next())
    {
        TV X=it.Location();

        T avg;
        if constexpr(TV::m == 2) {
            int a=it.Axis();
            T delta=grid.dX(1-a);
            TV v=TV::Axis_Vector(1-a);

            TV X1=X-0.5*v*delta;
            TV X2=X+0.5*v*delta;

            auto p = Quadrature_Integration<TV,T>(6, [&](VECTOR<T,2> w){
                TV X=X1*w(0)+X2*w(1);
                return interp.Interpolate_Face(grid,velocities,X);
                });
            avg = p(a);
        } else if constexpr(TV::m == 3) {
            TV dX=grid.dX;
            int a0=it.Axis();
            int a1=(it.Axis()+1)%3;
            int a2=(it.Axis()+2)%3;

            TV min_corner=-0.5*dX;TV max_corner=0.5*dX;
            RANGE<VECTOR<T, 2>> face({min_corner(a1),min_corner(a2)}, {max_corner(a1),max_corner(a2)});

            avg = 1./face.Edge_Lengths().Product()*Gauss_Quadrature_2D<T>([&](VECTOR<T, 2> offset){
                    TV Y=X;
                    Y(a1)+=offset(0);Y(a2)+=offset(1);
                    return interp.Interpolate_Face(grid,velocities,Y)(a0);
            },
            face,
            6);
        }

        T err=abs(avg-velocities(it.Full_Index()));
        if(err>max_err) max_err=err;
    }

    return max_err;
}

template<class TV>
void Run_Test()
{
    typedef VECTOR<int,TV::m> IV ;
    int res=10;
    GRID<TV> grid(IV()+res,RANGE<TV>::Unit_Box());
    ARRAY<T, FACE_INDEX<TV::m>> velocities(grid.Cell_Indices());

    auto field=Velocity_Field<TV>(0);
    for(FACE_ITERATOR<TV> it(grid);it.Valid();it.Next())
    {
        FACE_INDEX<TV::m> fi=it.Full_Index();
        velocities(fi)=field(it.Location())(it.Axis());
    }

    LINEAR_INTERPOLATION_FACE<TV,T> i1;
    CUBIC_INTERPOLATION_FACE<TV,T> i2;
    CUBIC_SPLINE_INTERPOLATION_FACE<TV,T> i3;
    DIV_FREE_INTERPOLATION<TV,T> i4;
    DIV_FREE_INTERPOLATION_FLUXES<TV,T> i5;
    DIV_FREE_BALSARA<TV,T> i6;

    std::array<INTERPOLATION_FACE<TV, T>*, 6> interps{&i1, &i2, &i3, &i4, &i5, &i6};
    std::array<std::string, 6> names{"Linear", "Cubic", "Cubic Spline", "Div Free C^0", "Div Free Flux-Preserving", "Balsara"};

    for(int i = 0; i < interps.size(); i++) 
    {
        INTERPOLATION_FACE<TV, T>* interp = interps[i];
        T max_err = Compute_Max_Flux_Error(grid,velocities,*interp);
        LOG::printf("%s: %d\n", names[i], max_err);
    }

}

int main(int argc,char** argv) {
    PARSE_ARGS parse_args(argc,argv);
    parse_args.Parse(true);

    LOG::printf("2D\n");
    Run_Test<VECTOR<double,2>>();

    LOG::printf("3D\n");
    Run_Test<VECTOR<double,3>>();

    return 0;
}
