//#####################################################################
// Copyright 2021, Craig Schroeder, Tamar Shinar.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Arrays/ARRAYS_ND.h>
#include <Core/Arrays/FACE_ARRAYS.h>
#include <Core/Boundaries/BOUNDARY.h>
#include <Core/Geometry_Particles/GEOMETRY_PARTICLES.h>
#include <Core/Grids/CELL_ITERATOR.h>
#include <Core/Grids/FACE_INDEX.h>
#include <Core/Grids/GRID.h>
#include <Core/Interpolation/CUBIC_INTERPOLATION_FACE.h>
#include <Core/Interpolation/CUBIC_SPLINE_INTERPOLATION_FACE.h>
#include <Core/Interpolation/GENERAL_INTERPOLATION_FACE.h>
#include <Core/Interpolation/LINEAR_INTERPOLATION_FACE.h>
#include <Core/Parsing/PARSE_ARGS.h>
#include <Core/Random_Numbers/RANDOM_NUMBERS.h>
#include <Core/Read_Write/FILE_UTILITIES.h>
#include <Core/Seeding/POISSON_DISK.h>
#include <Core/Utilities/VIEWER_DIR.h>
#include <Core/Vectors/VECTOR.h>

using namespace PhysBAM;

typedef float T;
typedef VECTOR<T,2> TV;
typedef VECTOR<int,TV::m> IV;


void Interpolate_Velocities_Div_Free(const GRID<TV>& grid,GEOMETRY_PARTICLES<TV>& particles,
    const ARRAY<T,FACE_INDEX<TV::m> >& mac_velocities)
{
    auto* valid=particles.Get_Array<bool>("valid");
    for(int p=0;p<particles.X.m;p++)
    {
        if(!(*valid)(p)) continue;
        TV X=particles.X(p);
        IV index=grid.Cell_I(X);
        TV dX=(X-grid.Node_X(index))*grid.one_over_dX;
        T x=dX.x,y=dX.y;
        auto u=[&mac_velocities,index](int i,int j)
            {return mac_velocities(FACE_INDEX<TV::m>(0,index+IV(i/2,(j-1)/2)));};
        auto v=[&mac_velocities,index](int i,int j)
            {return mac_velocities(FACE_INDEX<TV::m>(1,index+IV((i-1)/2,j/2)));};

        T U=-(T).125*(x-1)*(x-1)*(y-1)*u(-2,-1)
            +(T).125*(x-1)*(x-1)*u(-2,1)
            +(T).125*y*(x-1)*(x-1)*u(-2,3)
            +(T).125*(x*x-2)*(y-1)*u(0,-1)
            +(T).125*(x*x-2*x-1)*(y-1)*u(2,-1)
            -(T).125*x*x*(y-1)*u(4,-1)
            +(-(T).125*x*x+(T).25)*u(0,1)
            +(-(T).125*x*x+(T).25*x+(T).125)*u(2,1)
            +(T).125*x*x*u(4,1)
            -(T).125*y*(x*x-2)*u(0,3)
            -(T).125*y*(x*x-2*x-1)*u(2,3)
            +(T).125*x*x*y*u(4,3);

        T V=-(T).125*(y-1)*(y-1)*(x-1)*v(-1,-2)
            +(T).125*(y-1)*(y-1)*v(1,-2)
            +(T).125*x*(y-1)*(y-1)*v(3,-2)
            +(T).125*(y*y-2)*(x-1)*v(-1,0)
            +(-(T).125*y*y+(T).25)*v(1,0)
            -(T).125*x*(y*y-2)*v(3,0)
            +(T).125*(y*y-2*y-1)*(x-1)*v(-1,2)
            +((T).25*y-(T).125*y*y+(T).125)*v(1,2)
            -(T).125*x*(y*y-2*y-1)*v(3,2)
            -(T).125*y*y*(x-1)*v(-1,4)
            +(T).125*x*y*y*v(3,4)
            +(T).125*y*y*v(1,4);

        particles.V(p)=TV(U,V);
    }    
}

void Interpolate_Velocities_Div_Free_Tight(const GRID<TV>& grid,GEOMETRY_PARTICLES<TV>& particles,
    const ARRAY<T,FACE_INDEX<2> >& mac_velocities)
{
    auto* valid=particles.Get_Array<bool>("valid");
    for(int p=0;p<particles.X.m;p++)
    {
        if(!(*valid)(p)) continue;
        TV X=particles.X(p);
        IV index=grid.Cell_LC(X);
        TV dX=(X-grid.Cell_X(index))*grid.one_over_dX,&W=particles.V(p);
        for(int a=0;a<2;a++)
        {
            auto& U=mac_velocities.Component(a);
            const T* p=&U(index);
            int sa=U.stride(a),sb=U.stride(1-a);
            T ia=dX(a),ib=dX(1-a),ja=1-ia,jb=1-ib,u[3];
            for(int i=0;i<3;i++,p+=sa)
                u[i]=jb*p[0]+ib*p[sb];
            T v=ja*(ja*u[0]+ia*u[1])+ia*(ja*u[1]+ia*u[2]);
            W(a)=(T).5*(v+u[1]);
        }
    }    
}

namespace D3{
typedef VECTOR<T,3> TV;
typedef VECTOR<int,3> IV;
void Interpolate_Velocities_Div_Free_Tight(const GRID<TV>& grid,GEOMETRY_PARTICLES<TV>& particles,
    const ARRAY<T,FACE_INDEX<3> >& mac_velocities)
{
    auto* valid=particles.Get_Array<bool>("valid");
    for(int p=0;p<particles.X.m;p++)
    {
        if(!(*valid)(p)) continue;
        TV X=particles.X(p);
        IV index=grid.Cell_LC(X);
        TV dX=(X-grid.Cell_X(index))*grid.one_over_dX,&W=particles.V(p);
        for(int a=0;a<3;a++)
        {
            auto& U=mac_velocities.Component(a);
            const T* p=&U(index);
            int sa=U.stride(a),sb=U.stride((a+1)%3),sc=U.stride((a+2)%3);
            T ia=dX(a),ib=dX((a+1)%3),ic=dX((a+2)%3),ja=1-ia,jb=1-ib,jc=1-ic,u[3],w[2];
            for(int i=0;i<3;i++,p+=sa)
            {
                auto q=p;
                for(int j=0;j<2;j++,q+=sb)
                    w[j]=jc*q[0]+ic*q[sc];
                u[i]=jb*w[0]+ib*w[1];
            }
            T v=ja*(ja*u[0]+ia*u[1])+ia*(ja*u[1]+ia*u[2]);
            W(a)=(T).5*(v+u[1]);
        }
    }    
}
}

// w=quad(lin((A+B)/2,(D+E)/2),lin(B,E),lin((C+B)/2,(F+E)/2));
// w=quad((lin(A,D)+lin(B,E))/2,lin(B,E),(lin(C,F)+lin(B,E))/2);
// u0=lin(A,D);
// u1=lin(B,E);
// u2=lin(C,F);
// w=quad((u0+u1)/2,u1,(u2+u1)/2);
// w=quad((u0+u1)/2,(u1+u1)/2,(u2+u1)/2);
// w=quad(u0+u1,u1+u1,u2+u1)/2;
// w=(quad(u0,u1,u2)+u1)/2;

void Interpolate_Velocities(const GRID<TV>& grid,GEOMETRY_PARTICLES<TV>& particles,
    const ARRAY<T,FACE_INDEX<TV::m> >& mac_velocities,int interp_type)
{
    if(interp_type==0) return Interpolate_Velocities_Div_Free(grid,particles,mac_velocities);
    if(interp_type==4) return Interpolate_Velocities_Div_Free_Tight(grid,particles,mac_velocities);

    auto* valid=particles.Get_Array<bool>("valid");
    INTERPOLATION_FACE<TV,T>* interp=0;
    LINEAR_INTERPOLATION_FACE<TV,T> i1;
    CUBIC_INTERPOLATION_FACE<TV,T> i2;
    CUBIC_SPLINE_INTERPOLATION_FACE<TV,T> i3;
    if(interp_type==1) interp=&i1;
    else if(interp_type==2) interp=&i2;
    else interp=&i3;
    for(int p=0;p<particles.X.m;p++)
    {
        if(!(*valid)(p)) continue;
        particles.V(p)=interp->Interpolate_Face(grid,mac_velocities,particles.X(p));
    }
}

int main(int argc, char* argv[])
{
    PARSE_ARGS parse_args(argc,argv);
    std::string sim_dir;
    int particles_per_cell=4;
    int interp_type=0;
    int substeps=1;
    parse_args.Add("-sim",&sim_dir,"dir","simulation directory to trace particles on");
    parse_args.Add("-ppc",&particles_per_cell,"num","number of particles to seed per cell");
    parse_args.Add("-interp",&interp_type,"type","0=div-free, 1=bilinear, 2=quadratic, 3=cubic, 4=div-free-tight");
    parse_args.Add("-ss",&substeps,"steps","advection substeps");
    parse_args.Parse();

    VIEWER_DIR viewer_dir(sim_dir);
    STREAM_TYPE stream_type=STREAM_TYPE::type_float;
    GRID<TV> grid;
    viewer_dir.Set(0);
    Read_From_File(viewer_dir.Common_File("grid"),grid);
    LOG::printf("grid: %P\n",grid);
    
    T time=0;
    Read_From_File(viewer_dir.Frame_File("time"),time);
    ARRAY<bool,IV> psi_D;
    Read_From_File(viewer_dir.Frame_File("psi_D"),psi_D);

    ARRAY<T,FACE_INDEX<TV::m> > mac_velocities;
    ARRAY<T,FACE_INDEX<TV::m> > mac_velocities_ghost(grid.Cell_Indices(1));
    Read_From_File(viewer_dir.Frame_File("mac_velocities"),mac_velocities);
    BOUNDARY<TV,T> boundary;
    boundary.Set_Fixed_Boundary(false);
    boundary.Fill_Ghost_Faces(grid,mac_velocities,mac_velocities_ghost,1);

    RANDOM_NUMBERS<T> rand;
    rand.Set_Seed(123);
    GEOMETRY_PARTICLES<TV> particles;
    particles.Add_Array<bool>("valid");
    auto* valid=particles.Get_Array<bool>("valid");
    // for(CELL_ITERATOR<TV> it(grid);it.Valid();it.Next())
    // {
    //     if(psi_D(it.index)) continue;
    //     RANGE<TV> domain=grid.Cell_Domain(it.index);
    //     for(int i=0;i<particles_per_cell;i++)
    //     {
    //         int p=particles.Add_Element();
    //         rand.Fill_Uniform(particles.X(p),domain);
    //     }
    // }
    POISSON_DISK<TV> poisson_disk(grid.dX(0)/2);
    poisson_disk.Set_Distance_By_Volume(grid.Cell_Size()/particles_per_cell);

    ARRAY<TV> X;
    poisson_disk.Sample(rand,grid.domain,X);
    for(auto& x:X)
    {
        int p=particles.Add_Element();
        particles.X(p)=x;
    }
    valid->Fill(true);
    LOG::printf("particles: %P\n",particles.X.m);

    Interpolate_Velocities(grid,particles,mac_velocities,interp_type);
    Write_To_File(stream_type,viewer_dir.Frame_File("mpm_particles"),particles);
    
    while(viewer_dir.Find_Next_Directory(0,true))
    {
        T new_time=0;
        Read_From_File(viewer_dir.Frame_File("time"),new_time);
        T dt=new_time-time;

        for(int j=0;j<substeps;j++)
        {
            Interpolate_Velocities(grid,particles,mac_velocities,interp_type);
            for(int i=0;i<particles.X.m;i++)
            {
                particles.X(i)+=dt/substeps*particles.V(i);
                if(!grid.domain.Lazy_Inside(particles.X(i)))
                    (*valid)(i)=false;
            }
        }
            
        Read_From_File(viewer_dir.Frame_File("mac_velocities"),mac_velocities);
        boundary.Fill_Ghost_Faces(grid,mac_velocities,mac_velocities_ghost,1);
        Read_From_File(viewer_dir.Frame_File("psi_D"),psi_D);

        Interpolate_Velocities(grid,particles,mac_velocities,interp_type);
        Write_To_File(stream_type,viewer_dir.Frame_File("mpm_particles"),particles);

        time=new_time;
        LOG::printf("%P\n",viewer_dir.Frame_As_String());
    }
 
    
    return 0;
}

