#include <Core/Arrays/ARRAY.h>
#include <Core/Grids/FACE_INDEX.h>
#include <Core/Grids/FACE_ITERATOR.h>
#include <Core/Grids/GRID.h>
#include <Core/Parsing/PARSE_ARGS.h>
#include <Core/Random_Numbers/RANDOM_NUMBERS.h>
#include <Core/Vectors/VECTOR.h>
#include "INTERPOLATION_SCHEMES.h"
#include "VELOCITY_FIELDS.h"

using namespace PhysBAM;

typedef double RW;
typedef double T;

template<class TV>
void Test_Resolution(int res,const ARRAY<TV>& positions,
    const ARRAY<TV>& exact_velocities,int velocity_field,
    ARRAY<INTERPOLATION_SCHEMES<TV>>& interps)
{
    typedef VECTOR<int,TV::m> IV;
    GRID<TV> grid(IV()+res,RANGE<TV>::Unit_Box());

    ARRAY<T,FACE_INDEX<TV::m>> velocity(grid.Cell_Indices());
    ARRAY<T,FACE_INDEX<TV::m>> averaged_velocity(grid.Cell_Indices());
    auto field=Velocity_Field<TV>(velocity_field);
    auto avg_field=Velocity_Field_Average<TV>(velocity_field);
    for(FACE_ITERATOR it(grid); it.Valid(); it.Next()) 
    {
        auto fi=it.Full_Index();
        TV X=it.Location();

        velocity(fi)=field(X)(fi.axis);
        if(avg_field) averaged_velocity(fi)=avg_field(X,grid.dX.Max())(fi.axis);
    }

    LOG::printf("%P",res);
    for(int i=0; i<interps.m; i++) 
    {
        T max_err=0;;
        ARRAY<T,FACE_INDEX<TV::m>>& data=velocity;
        if(i == 9) data=averaged_velocity;

        for(int p=0; p<positions.m; p++) 
        {
            TV V=interps(i).interp->Interpolate_Face(grid,data,positions(p));
            T err=(V - exact_velocities(p)).Magnitude();
            max_err=max(max_err,err);
        }

        LOG::printf(" %P",max_err);
    }
    LOG::printf("\n");
}

template<int d>
void Run(PARSE_ARGS& parse_args)
{
    typedef VECTOR<T,d> TV;
    typedef VECTOR<int,TV::m> IV;
    int velocity_field=1;
    int refine=8;
    bool names_only=false;
    bool curl=false;
    parse_args.Add("-vel",&velocity_field,"velocity field","which velocity field to use (0-3)");
    parse_args.Add("-refine",&refine,"refinement","refinement level");
    parse_args.Add("-names",&names_only,"print names and exit");
    parse_args.Add("-curl",&curl,"process curl-free schemes");
    parse_args.Parse();

    auto list=Interpolation_Schemes<TV>();
    ARRAY<INTERPOLATION_SCHEMES<TV>> keep;
    for(auto s:list)
        if((curl&&!s.is_div) || (!curl&&!s.is_curl))
            keep.Append(s);
    
    if(names_only)
    {
        LOG::printf("res");
        for(auto s:keep)
            LOG::printf(" %s",s.name);
        LOG::printf("\n");
        return;
    }

    RANDOM_NUMBERS<T> rand(123);

    ARRAY<TV> positions(1000000);
    rand.Fill_Uniform(positions,TV(),TV()+1);

    ARRAY<TV> velocities(positions.m);
    auto field=Velocity_Field<TV>(velocity_field);
    for(int p=0; p<positions.m; p++) 
        velocities(p)=field(positions(p));

    Test_Resolution(refine,positions,velocities,velocity_field,keep);
}

int main(int argc,char* argv[])
{
    PARSE_ARGS parse_args(argc,argv);
    int dim=2;
    parse_args.Add("-dim",&dim,"dim","dimension to test");
    parse_args.Parse(true);

    if(dim==2) Run<2>(parse_args);
    else Run<3>(parse_args);
    return 0;
}
